#ifndef BASE_MESSAGE_H
#define BASE_MESSAGE_H

#include <string>

//called 'message' because it was originally designed for network code
//static buffers with read/write << >> operations implemented at the binary level

//message

//TODO - make a ReadMsg class for the recv() function
//and then make a MsgTable class for net-encoding/decoding structures, RLE data, etc

//stores bytestreams to be sent out to the network
//this does not null-terminate! nor does it return an error (typically) upon overflow
template<int _capacity = 256>
class Message {
public:
	char data[_capacity];
	int seek;		//the current reading index of our message
	int size;		//the current size of our message
	bool overflow;	//whether our size was asked to surpass the capacity -- or whether the seek was asked to surpass the size -- both right now
	
	Message() {
		reset();
	}
	
	void reset(void) {
		seek = 0;
		size = 0;
		overflow = false;
		//no need to memset - we'll just fill it as size increases
	}
	
	int capacity(void) { return _capacity; }
	
//writing operations . . .
	
	//return true/false whether this many bytes will fit
	bool fits(int extra) {
		return size + extra <= _capacity;
	}
	
	//return true if 'extra' bytes will fit in our data buffer
	//return false and sets overflow flag otherwise
	bool mustfit(int extra) {
		if (fits(extra)) return true;
		overflow = true;
		return false;
	}
		
	void step(int extra) {
		size += extra;
		if (size > _capacity) {
			size = _capacity;
			overflow = true;	//shouldnt be a need to set it here since mustfit() should always be called beforehand, and it sets the overflow bit
		}
	}

	//raw data copy - no string term
	void pushraw(const void *b, int len) {
		if (mustfit(len)) {
			memcpy(data+size, b, len);
			step(len);
		}
	}
	
	//byte operators
	//creates operator<< for each type
	//and creates a "writec", "writeuc", etc, for each type
	//i suppose i should stick to opengl primitive suffix conventions .. . 
#define WRITE_OP(__type,__prefix)		\
	Message &operator<<(__type b) {		\
		if (mustfit(sizeof(__type))) {	\
			*(__type *)(data+size) = b;	\
			step(sizeof(__type));		\
		}								\
		return *this;					\
	}									\
	void write##__prefix(__type b) {	\
		*this << b;						\
	}

	WRITE_OP(char,b)
	WRITE_OP(unsigned char,ub)
	WRITE_OP(short,s)
	WRITE_OP(int,i)
	WRITE_OP(float,f)

	//include the null-terminator for strings.  plus one the strlen for the null-terminator
	Message &operator<<(const char *b) 	{ pushraw(b, strlen(b)+1); return *this; }
	
	//for strings - works like const char *'s
	Message &operator<<(const std::string &s) { pushraw(s.c_str(), s.length()+1); return *this; }

	//for pushing entire messages onto it
	template<int T>
	Message &operator<<(const Message<T> &m) { pushraw(m.data, m.size); return *this; }
	
//reading operations

	//return true/false whether this many bytes will fit
	bool fitseek(int extra) {
		return seek + extra <= size;
	}
	
	//return true if 'extra' bytes will fit in our data buffer
	//return false and sets overflow flag otherwise
	bool mustfitseek(int extra) {
		if (fitseek(extra)) return true;
		overflow = true;
		return false;
	}
		
	void stepseek(int extra) {
		seek += extra;
		if (seek > size) {
			seek = size;
			overflow = true;	//shouldnt be a need to set it here since mustfit() should always be called beforehand, and it sets the overflow bit
		}
	}

#define READ_OP(__type,__prefix)			\
	__type	read##__prefix(void) { 			\
		__type b = 0; 						\
		if (mustfitseek(sizeof(__type))) {	\
			b = *(__type *)(data+seek);		\
			stepseek(sizeof(__type));		\
		}									\
		return b;							\
	}
	
	READ_OP(char,b)
	READ_OP(short,s)
	READ_OP(int,i)
	READ_OP(float,f)
};

#endif //BASE_MESSAGE_H