#ifndef BASE_TEXTURE_H
#define BASE_TEXTURE_H

#include <assert.h>
#include <vector>

#include "system/gl.h"
#include "system/glu.h"
//#ifdef WIN32
//#include <gl/glaux.h>	//GLAUX SUCKS!
//#endif

#include "image.h"
#include "opengl.h"

/**
 * For all the generic OpenGL texture stuff not associated with any one bound texture.
 * Maybe some texture loading stuff too ...
 */
class TextureSystem {
protected:
	GLint maxUnits;
public:

	TextureSystem() : maxUnits(0) {}

	//init texture system states / gathers static information
	void init();

	int getMaxUnits() { return maxUnits; }
};
extern TextureSystem textureSystem;

/**
 * This class represents a texture object stored in OpenGL.
 * It is used as a wrapper for all OpenGL functionality, mostly to save a programmer some typing.
 */
class Texture {
protected:

	/**
	 * The index generated by glGenTextures
	 */
	GLuint texture_id;

	/**
	 * Either TEXTURE_1D, TEXTURE_2D, TEXTURE_CUBE_MAP_ARB, or TEXTURE_3D
	 * Currently only TEXTURE_1D, TEXTURE_2D, and TEXTURE_CUBE_MAP_ARB are supported.
	 */
	int target;

	/**
	 * Clear all variables and restore this object to its default state
	 */
	void reset();

	/**
	 * extracts the target from a 1D buffer
	 * - returns NULL if it doesnt need extraction
	 */
	unsigned char *extract1DBuffer(int *_components, int width, int format, int type, const unsigned char *ptr, int colskip = 0);

	/**
	 * extracts the target from a 2D buffer
	 * - returns NULL if it doesnt need extraction
	 */
	unsigned char *extract2DBuffer(int *_components, int width, int height, int format, int type, const unsigned char *ptr, int colskip = 0, int rowskip = 0);

public:

	Texture() {reset();}
	~Texture() {unload();}

	/**
	 * This safely unloads all loaded resources from the object.
	 * Specifically this will free the allocated texture_id value from OpenGL.
	 */
	void unload();

	/**
	 * the function used to initialize textures in OpenGL
	 * and to generate this's texture_id
	 * while we're here, give it a target too
	 */
	void initTextureID(int target);

	/**
	 * initialization for cube maps
	 */
	bool initCubeMap();

	/**
	 * Shell function for gluBuild1DMipmaps
	 * if colskip < the size of a pixel then it will assume it is equal to the size of a pixel
	 * if colskip > the size of a pixel then it will allocate a new buffer for uploading the data
	 * TODO - have mipmapping optional
	 * (mipmap generation is unneeded when generateMipmapSGIS is set)
	 */
	bool build1DMipmaps(int width, int format, int type, const unsigned char *ptr, int colskip = 0);

	/**
	 * Shell function for gluBuild2DMipmaps
	 * TODO - have mipmapping optional
	 * (mipmap generation is unneeded when generateMipmapSGIS is set)
	 */
	bool build2DMipmaps(int width, int height, int format, int type, const unsigned char *ptr, int colskip = 0, int rowskip = 0);
	
	//like build2DMipmaps but...
	//1) no mipmaps
	//2) assumes stuff is comapct
	//3) assumes texture exists
	//basically used for fast access / realtime updates
	void updateData2D(int width, int height, int internal_format, int external_format, void *src);

	/**
	 * creates an empty 2d texture
	 * leaving componets the default causes createEmpty to look up its correct value depending on the format
	 */
	bool createEmpty(int width, int height, int format, int type, int components = 0);

	/**
	 * creates an empty cube map with all six surfaces of the specified dimensions
	 * with no texture data.
	 */
	bool createEmptyCubeMap();

	/**
	 * side = GL_TEXTURE_CUBE_MAP_{POSITIVE|NEGATIVE}_{X|Y|Z}_ARB constant
	 * all else - is passed along to glTexImage2D
	 * TODO - should 'side' be based upon GL_TEXTURE_CUBE_MAP_POSTIIVE_X_ARB + 0..5
	 *			or should 'side' just be 0..5?
	 * TODO - have mipmapping optional
	 */
	bool createCubeMapSide(int side, int width, int height, int format, int type, const unsigned char *ptr, int colskip = 0, int rowskip = 0);

	/**
	 * uses texSubImage for faster upload than creating a fully new texture would
	 * TODO - have mipmapping optional
	 */
	bool updateCubeMapSide(int side, int width, int height, int format, int type, const unsigned char *ptr, int colskip = 0, int rowskip = 0);

	/**
	 * Shell function for auxDIBImageLoad - passes data along to gluBuild2DMipmaps
	 */
	bool load2D(const char *filename);

	/**
	 *  same as above but for any arbitrary Image
	 */
	bool loadImage(const Image &img);

	/**
	 * Shell function for auxDIBImageLoad/gluBuild2DMipmaps
	 * combined with GL_TEXTURE_CUBE_MAP_ARB and GL_TEXTURE_CUBE_MAP_{NEGATIVE|POSITIVE}_{X|Y|Z}_ARB
	 * fails if the GL_ARB_texture_cube_map extension was not detected
	 */
	bool loadCubeMap(
		const char *filename0,
		const char *filename1,
		const char *filename2,
		const char *filename3,
		const char *filename4,
		const char *filename5);

	/**
	 * Shell function for glBindTexture
	 * This also enables the texture's target
	 */
	void bind() const {
		//disable any previous targets (to prevent any from overriding us)
		unbind();
		//enable the target we're looking for
		glEnable(target);
		//and bind the texture
		glBindTexture(target, texture_id);
	}

	void bind(int unit) const { setUnit(unit); bind(); }

	//multitexture unit setting / resetting
	static void setUnit(int unit) {
		if (found_GL_ARB_multitexture) {
			glActiveTextureARB(GL_TEXTURE0_ARB + unit);
		}
	}

	static void resetUnit() { setUnit(0); }

	/**
	 * Disables all glEnable(GL_TEXTURE***)'s
	 */
	static void unbind() {
		glDisable(GL_TEXTURE_1D);
		glDisable(GL_TEXTURE_2D);
		if (found_GL_ARB_texture_cube_map) {
			glDisable(GL_TEXTURE_CUBE_MAP_ARB);
		}
	}

	static void unbindUnit(int unit) { setUnit(unit); unbind(); }

	//specific target unbinds ... 
	void unbindMe() { if (target) glDisable(target); }
	void unbindMe(int unit) { setUnit(unit); unbindMe(); }

	static void unbindAll() {
		//do it backwards so
		//1) we only poll getMaxUnits() once
		//2) we end with unit0 bound
		for (int i = textureSystem.getMaxUnits()-1; i >= 0; i--) {
			unbindUnit(i);
		}
	}

	/**
	 * Shell function for glTexParameteri
	 * for the GL_WRAP_S and GL_WRAP_T parameters
	 *
	 * TODO - add support for GL_WRAP_R and GL_WRAP_Q
	 */
	void setWrap1D(int mode_s) {
		bind();
		glTexParameteri(target, GL_TEXTURE_WRAP_S, mode_s);
	}

	void setWrap2D(int mode_s, int mode_t) {
		bind();
		glTexParameteri(target, GL_TEXTURE_WRAP_S, mode_s);
		glTexParameteri(target, GL_TEXTURE_WRAP_T, mode_t);
	}

	/**
	 * Shell function for glTexParameteri's GL_TEXTURE_MIN_FILTER and GL_TEXTURE_MAX_FILTER
	 */
	void setFilter(int mag, int min) {
		bind();
		glTexParameteri(target, GL_TEXTURE_MAG_FILTER, mag);
		glTexParameteri(target, GL_TEXTURE_MIN_FILTER, min);
	}

	/**
	 * Set to 'true' to enable glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, enable)
	 *	only works if GL_SGIS_generate_mipmap is found
	 *
	 * TODO - add support for glTexParameteri(target, GL_GENERATE_MIPMAP_HINT_SGIS, GL_{FASTEST|NICEST});
	 * TODO - add support for the GL_SGIS_texture_lod extension
	 * TODO - add a new function for glGenerateMipmapEXT (which works better)
	 */
	void generateMipmapSGIS(bool enable) {
		if (found_GL_SGIS_generate_mipmap) {
			bind();
			glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, enable);
		}
	}

	/**
	 * specifies the quality of the mipmaps generated
	 * mode = either GL_NICEST or GL_FASTEST
	 */
	void generateMipmapHintSGIS(int mode) {
		if (found_GL_SGIS_generate_mipmap) {
			bind();
			glTexParameteri(target, GL_GENERATE_MIPMAP_HINT_SGIS, mode);
		}
	}

	//read-only
	int getTextureID() const {
		return texture_id;
	}

	//all require bind() to be called
	int getWidth() const;
	int getHeight() const;
	int getFormat() const;
	void *getData(int level = 0, int format = GL_RGB, int type = GL_UNSIGNED_BYTE) const;
};

#endif