#ifndef BASE_VIEWPORT_H
#define BASE_VIEWPORT_H

#include <assert.h>
#include "view.h"
#include "scene.h"
#include "defs/defs.h"
#include "math/vec.h"

/**
 * The Viewport class represents a 2D region on the window surface.
 * Viewports reference View objects which represent where within a Scene
 * to render our display.
 * Viewports also reference Scene objects which describe what data to display,
 * yet in the future the Scene pointer may be moved to the View class.
 * Viewports also hold frustum information of the viewing volume within
 * the scene.
 */
class Viewport {

protected:
	/**
	 * The position and size of the viewport -
	 * used for glViewport() calls
	 */
	box2i	area;

	/**
	 * The object used to view the Scene through this Viewport
	 */
	View	*view;

	/**
	 * The frustum of the Viewport's View object within the Viewport's Scene
	 */
	Frustum	frustum;

	/**
	 * Clear all variables and restore this object to its default state
	 */
	void	reset();

public:

	/**
	 * Construct an initial Viewport
	 */
	Viewport() {reset();}

	/**
	 * Render this Viewport's View
	 */
	void render();

	/**
	 * glViewport, glClear, and stencil enable
	 * TODO - make stencil optional
	 */
	void setupViewport();

	/**
	 * clear
	 */
	void clear();

	/** 
	 * setup the frustm by the curret view
	 * separated for ext frustm calc's sake (its proctect otherwise)
	 */
	void setupFrustum();

	/**
	 * stencil disable
	 */
	void shutdownViewport();

	/**
	 * Returns 'true' if the 2D point is within the viewport, 'false' otherwise
	 * TODO - give the Viewport a rectangle class object representing <x,y,width,height>
	 * and give the user access to the rectangle to test whether it contains a point.
	 */
	bool containsPoint(const vec2i &v) {
		return area.containsE(v);
		//return (s >= x) && (s < x + width) && (t >= y) && (t < y + height);
	}

	/**
	 * Returns the position, in modelview space, of the 2D point, in screen coordinates
	 */
	vec3f getPixelPos(const vec2i &v) const;
	
	/**
	 * Returns the line that traces from the origin through the pixel position
	 */
	line3f getPixelTraceLine(const vec2i &p) const {
		assert(view);
		return line3f( view->getPos(),  getPixelPos(p) );
	}


	//// inline variable access:


	/**
	 * Sets the 2D region for the viewport
	 */
	void set(int _x, int _y, int _w, int _h) {
		if (_w < 1) _w = 1;
		if (_h < 1) _h = 1;
		area.set(_x, _y, _x+_w, _y+_h);
	}

	/**
	 * Returns individual parameters of the viewport's region
	 */
	int getX() { return area.min.x; }
	int getY() { return area.min.y; }
	int getWidth() const { return area.max.x - area.min.x; }
	int getHeight() const { return area.max.y - area.min.y; }

	/**
	 * Set or get the view pointer
	 */
	void setView(View *view) { this->view = view; }
	View *getView() const { return view; }

	/**
	 * returns our aspect ratio - currently height/width (FIX THIS)
	 */
	float getAspectRatio() const { return (float)(area.max.y - area.min.y) / (float)(area.max.x - area.min.x); }

	/**
	 * return a pointer to the frustum object.
	 */
	REPEAT_WITH_CONST( Frustum &getFrustum() , return frustum; )
	
	//some extra stuffs:
	
	//whether to use scissor tests with the viewport clear command
	//glScissor - off by default
	bool useScissor;
	
	//glClearColor - off by default
	bool useClearColor;
	vec4f clearColor;
	
	//glClear - color, depth, and stencil by default, on by default
	bool useClear;
	int clearFlags;
};

#endif
